/*
  [Digi] A new id3 module to support characters encodings.
*/

/* id3.c
 * - Functions for id3 tags in ices
 * Copyright (c) 2000 Alexander Havng
 * Copyright (c) 2001-3 Brendan Cully <brendan@icecast.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "definitions.h"
#include "metadata.h"
#include <iconv.h>
#include <debug.h>

extern ices_config_t ices_config;

/* Local definitions */
typedef struct {
  unsigned char major_version;
  unsigned char minor_version;
  unsigned char flags;
  size_t len;

  int found; // [0..3], 3 when "artist", "title" and "album" tags found.

  unsigned int pos;
} id3v2_tag;

typedef struct {
  size_t frame_len;
  const char* artist_tag;
  const char* title_tag;
  const char* album_tag;
} id3v2_version_info;

static id3v2_version_info vi[5] = {
  {  6, "TP1",  "TT2", "TAL"  },
  {  6, "TP1",  "TT2", "TAL"  },
  {  6, "TP1",  "TT2", "TAL"  },
  { 10, "TPE1", "TIT2", "TALB" },
  { 10, "TPE1", "TIT2", "TALB" }
};

#define ID3V2_FLAG_UNSYNC (1 << 7)
#define ID3V2_FLAG_EXTHDR (1 << 6)
#define ID3V2_FLAG_EXPHDR (1 << 5)
#define ID3V2_FLAG_FOOTER (1 << 4)

#define ID3V2_FRAME_LEN(tagp) (vi[(tagp)->major_version].frame_len)
#define ID3V2_ARTIST_TAG(tagp) (vi[(tagp)->major_version].artist_tag)
#define ID3V2_TITLE_TAG(tagp) (vi[(tagp)->major_version].title_tag)
#define ID3V2_ALBUM_TAG(tagp) (vi[(tagp)->major_version].album_tag)

/* Private function declarations */
static int id3v2_read_exthdr(input_stream_t* source, id3v2_tag* tag);
ssize_t id3v2_read_frame(input_stream_t* source, id3v2_tag* tag);
static int id3v2_skip_data(input_stream_t* source, id3v2_tag* tag, size_t len);
static int id3v2_decode_synchsafe(unsigned char* synchsafe);
static int id3v2_decode_synchsafe3(unsigned char* synchsafe);
static int id3v2_decode_unsafe(unsigned char* in);
static int id3v2_convert_unicode(char *str, int len);

/* Global function definitions */

static int read_id3v1_item(int fd, int iMDItem)
{
  char       acBuf[31];
  char       *pcEnd;

  if ( read( fd, acBuf, 30 ) != 30 )
    return 0;

  pcEnd = strchr( acBuf, '\0' );
  while( ( pcEnd > acBuf ) && ( *(pcEnd - 1) == ' ' ) )
    pcEnd--;

  ices_metadata_set_item( iMDItem, acBuf, strlen( acBuf ),
                          ices_config.pm.id3v1_enc );
  return 1;
}

void ices_id3v1_parse(input_stream_t* source)
{
  off_t      pos;
  char       acBuf[64];

  if ( source->filesize == 0 )
    return;

  pos = lseek( source->fd, 0, SEEK_CUR );

  lseek( source->fd, -128, SEEK_END );

  if ( ( read(source->fd, acBuf, 3) == 3 ) &&
       ( strncmp( acBuf, "TAG", 3 ) == 0 ) )
  {
    /* Don't stream the tag */
    source->filesize -= 128;

    if ( !read_id3v1_item( source->fd, MDITEM_TITLE ) ||
         !read_id3v1_item( source->fd, MDITEM_ARTIST ) ||
         !read_id3v1_item( source->fd, MDITEM_ALBUM ) )
    {
      ices_log( "Error reading ID3v1 item" );
    }
    else if ( ices_config.verbose )
    {
      ices_metadata_get_item( MDITEM_TITLE, acBuf, sizeof(acBuf), NULL );
      ices_log_debug( "ID3v1: Title: %s", acBuf );
      ices_metadata_get_item( MDITEM_ARTIST, acBuf, sizeof(acBuf), NULL );
      ices_log_debug( "ID3v1: Artist: %s", acBuf );
      ices_metadata_get_item( MDITEM_ALBUM, acBuf, sizeof(acBuf), NULL );
      ices_log_debug( "ID3v1: Album: %s", acBuf );
    }
  }

  lseek( source->fd, pos, SEEK_SET );
}

void ices_id3v2_parse(input_stream_t* source)
{
  unsigned char buf[1024];
  id3v2_tag tag;
  size_t remaining;
  ssize_t rv;

  if (source->read(source, buf, 10) != 10) {
    ices_log("Error reading ID3v2");

    return;
  }

  tag.found = 0;
  tag.pos = 0;
  tag.major_version = *(buf + 3);
  tag.minor_version = *(buf + 4);
  tag.flags = *(buf + 5);
  tag.len = id3v2_decode_synchsafe(buf + 6);
  ices_log_debug("ID3v2: version %d.%d. Tag size is %d bytes.",
           tag.major_version, tag.minor_version, tag.len);
  if (tag.major_version > 4) {
    ices_log_debug("ID3v2: Version greater than maximum supported (4), skipping");
    id3v2_skip_data(source, &tag, tag.len);

    return;
  }

  if ((tag.major_version > 2) &&
      (tag.flags & ID3V2_FLAG_EXTHDR) && id3v2_read_exthdr(source, &tag) < 0) {
    ices_log("Error reading ID3v2 extended header");

    return;
  }

  remaining = tag.len - tag.pos;
  if ((tag.major_version > 3) && (tag.flags & ID3V2_FLAG_FOOTER))
    remaining -= 10;

  while( ( remaining > ID3V2_FRAME_LEN(&tag) ) && ( tag.found < 3 ) )
  {
    if ( (rv = id3v2_read_frame(source, &tag)) < 0 )
    {
      ices_log("Error reading ID3v2 frames, skipping to end of ID3v2 tag");
      break;
    }

    /* found padding */
    if (rv == 0)
      break;

    remaining -= rv;
  }

  remaining = tag.len - tag.pos;
  if ( remaining )
    id3v2_skip_data(source, &tag, remaining);
}

static int id3v2_read_exthdr(input_stream_t* source, id3v2_tag* tag) {
  char hdr[6];
  size_t len;

  if (source->read(source, hdr, 6) != 6) {
    ices_log("Error reading ID3v2 extended header");

    return -1;
  }
  tag->pos += 6;

  len = id3v2_decode_synchsafe((unsigned char*)hdr);
  ices_log_debug("ID3v2: %d byte extended header found, skipping.", len);

  if (len > 6)
    return id3v2_skip_data(source, tag, len - 6);
  else
    return 0;
}

ssize_t id3v2_read_frame(input_stream_t* source, id3v2_tag* tag)
{
  char       hdr[10];
  size_t     len, len2;
  ssize_t    rlen;
  char*      buf;
  int        metadata_field;

  // Textual frames encoding byte values.
  static char*  id3v2_enc_list[4] =
                       { "ISO-8859-1", "UCS-2", "UTF-16BE", "UTF-8" };

  if ( source->read(source, hdr, ID3V2_FRAME_LEN(tag)) != ID3V2_FRAME_LEN(tag) )
  {
    ices_log("Error reading ID3v2 frame");
    return -1;
  }
  tag->pos += ID3V2_FRAME_LEN(tag);

  if ( hdr[0] == '\0' )
    return 0;

  if ( tag->major_version < 3 )
  {
    len = id3v2_decode_synchsafe3((unsigned char*)(hdr + 3));
    hdr[3] = '\0';
  }
  else if (tag->major_version == 3)
  {
    len = id3v2_decode_unsafe((unsigned char*)(hdr + 4));
    hdr[4] = '\0';
  }
  else
  {
    len = id3v2_decode_synchsafe((unsigned char*)(hdr + 4));
    hdr[4] = '\0';
  }

  if ( len > (tag->len - tag->pos) )
  {
    ices_log( "Error parsing ID3v2 frame header: Frame too large (%d bytes)",
              len );
    return -1;
  }

  if ( strcmp(hdr, ID3V2_TITLE_TAG(tag)) == 0 )
    metadata_field = MDITEM_TITLE;
  else if ( strcmp(hdr, ID3V2_ARTIST_TAG(tag)) == 0 )
    metadata_field = MDITEM_ARTIST;
  else if ( strcmp(hdr, ID3V2_ALBUM_TAG(tag)) == 0 )
    metadata_field = MDITEM_ALBUM;
  else
    metadata_field = -1;

  if ( metadata_field != -1 )
  {
    if ( !(buf = debugMAlloc(len + 1)) )
    {
      ices_log( "Error allocating memory while reading ID3v2 frame" );
      return -1;
    }

    len2 = len;
    while( len2 )
    {
      if ( (rlen = source->read(source, buf, len2)) < 0 )
      {
        ices_log( "Error reading ID3v2 frame data" );
        debugFree(buf);

        return -1;
      }
      tag->pos += rlen;
      len2 -= rlen;
    }

    // Store metadata field's value and encoding.

    // I have mp3s with zero padding in tags...
    len2 = len - 1;
    while( len2 > 0 && buf[len2 - 1] == '\0' )
      len2--;

    // For stupid situation when ID3v2 tag contains national characters and
    // has type 0 (ISO-8859-1 or Latin1) - we use configured for id3v1 charset.
    if ( ices_config.pm.id3v1_enc != NULL )
    {
      int        iIdx;
      char       *pch;

      id3v2_enc_list[0] = "ISO-8859-1";
      for( iIdx = 0, pch = &buf[1]; iIdx < len2; iIdx++, pch++ )
      {
        if ( (*pch & 0x80) != 0 )
        {
          id3v2_enc_list[0] = ices_config.pm.id3v1_enc;
          break;
        }
      }
    }

    ices_metadata_set_item( metadata_field, &buf[1], len2,
                            buf[0] >= ( sizeof(id3v2_enc_list) / sizeof(char *) )
                                       ? id3v2_enc_list[0]
                                       : id3v2_enc_list[ buf[0] ] );
    tag->found++;

    if ( ices_config.verbose )
    {
      static const char *acMetadataFldName[] = { "Artist", "Title", "Album" };

      ices_metadata_get_item( metadata_field, buf, len + 1, NULL );
      ices_log_debug( "ID3v2: %s found: %s", acMetadataFldName[metadata_field],
                      buf );
    }

    debugFree( buf );
  }
  else if ( id3v2_skip_data(source, tag, len) )
    return -1;

  return len + ID3V2_FRAME_LEN(tag);
}

static int id3v2_skip_data(input_stream_t* source, id3v2_tag* tag, size_t len)
{
  char* buf;
  ssize_t rlen;

  if (!(buf = debugMAlloc(len))) {
    ices_log("Error allocating memory while skipping ID3v2 data");

    return -1;
  }

  while (len) {
    if ((rlen = source->read(source, buf, len)) < 0) {
      ices_log("Error skipping in ID3v2 tag.");
      debugFree(buf);

      return -1;
    }
    tag->pos += rlen;
    len -= rlen;
  }

  debugFree(buf);

  return 0;
}

static int id3v2_decode_synchsafe(unsigned char* synchsafe)
{
  int res;

  res = synchsafe[3];
  res |= synchsafe[2] << 7;
  res |= synchsafe[1] << 14;
  res |= synchsafe[0] << 21;

  return res;
}

static int id3v2_decode_synchsafe3(unsigned char* synchsafe) {
  int res;

  res = synchsafe[2];
  res |= synchsafe[1] << 7;
  res |= synchsafe[0] << 14;

  return res;
}

/* id3v2.3 badly specifies frame length */
static int id3v2_decode_unsafe(unsigned char* in)
{
  int res;

  res = in[3];
  res |= in[2] << 8;
  res |= in[1] << 16;
  res |= in[0] << 24;

  return res;
}
